;;;; -*-Mode:LISP; Package:LISP; Base:10; Syntax:ISLISP -*-
;;;; Date:      2006/10/06
;;;; Title:     defclass.lsp
;;;; Author:    C. Jullien
;;;; RCS:       "$Id: defclass.lsp,v 1.6 2007/01/01 10:11:01 jullien Exp $"

;;; DEFCLASS package.

;;; This  module implements the defclass macro that expand the object
;;; system  with  user classes.  All user objects inherit directly or
;;; indirectly  from  <standard-object>  class.  This  implementation
;;; fully  conforms  to  ISO/IEC  13816 ISLISP Standard.  Most of the
;;; code is done in C (class.c)

(require 'setf)

(in-package "openlisp")

(defpackage "defclass"
   (:use    "openlisp")
   (:export *initarg-as-keywords*
            defclass slot-unbound slot-missing
            %build-class
            %true-initform
            %false-initform
            %one-initform
            %zero-initform)
   (:import-from "openlisp"
            *compile-file* *compiler-struct-ht*))

(in-package "defclass")

(defglobal *initarg-as-keywords* ())    ;; ISLISP standard (CLtL is t)

;;; The DEFCLASS macro.

(defmacro defclass (name sc-list slot-spec &rest class-opt)
   ;; Define a new class,  just a wrapper to the real %expand-defclass
   ;; function
   (%expand-defclass name sc-list slot-spec class-opt))

(defgeneric slot-unbound (cl instance slot)
   ;; Default method for unbound slots.
   (:method ((cl <standard-class>) (instance <object>) (slot <object>))
            (error "unbound slot ~a for a~%" slot instance)))

(defgeneric slot-missing (cl instance slot caller)
   ;; Default method for unbound slots.
   (:method ((cl       <standard-class>)
             (instance <object>)
             (slot     <object>)
             (caller   <symbol>))
            (error "~a : slot missing ~a : ~a~%" caller slot instance)))

;;; Helper functions

(defconstant *slot-option-keywords*
             ;; slot keywords that define generic functions
             '(:accessor :reader :writer :boundp))

(defun %expand-defclass (name sc-list slot-spec class-opt)
   ;; Check the name
   (unless (symbolp name)
           (error "defclass: illegal-class-name ~a~%" name))
   ;; Check super-classes arguments
   (setq sc-list (mapcar (lambda (x)
                            (if (not (symbolp x))
                                (error "defclass: bad-superclass-list ~a~%" x)
                                x))
                         sc-list))
   ;; Check slots arguments
   (unless (every #'consp slot-spec)
           (error "defclass: bad-slot-descriptions ~a~%" slot-spec))
   (let ((initform  ())
         (initarg   ())
         (fields    ())
         (abstractp ())
         (options   ())
         (slots     ()))
        (setq slots     (%compute-slot-declarations sc-list slot-spec))
        (setq abstractp (some (lambda (opt)
                                 (if (and (consp opt)
                                          (eq (car opt) :abstractp))
                                     (cadr opt)
                                     (error "DEFCLASS: wrong option: '~s'~%"
                                            opt)))
                              class-opt))
        (setq fields    (mapcar (lambda (x)
                                   (if (consp x)
                                       (car x)
                                       x))
                                slots))
        (setq options   (let ((i -1))
                             (mapcar (lambda (x)
                                        (%compute-slot-options name x (incf i)))
                                     slots)))
        (setq initform  (cons 'list (mapcar #'car options)))
        (setq initarg   (mapcar #'cadr options))
        (setq slots     (apply #'append (mapcar #'caddr options)))
#+*initarg-as-keywords*
        (setq keywords  (mapcar (lambda (x)
                                   (when x `(defconstant ,x ',x)))
                                initarg))
        `(progn
               (%build-class ',name
                             ',(or sc-list '(<standard-object>))
                             ,abstractp
                             (create-class-info
                                     ',fields
                                     ,initform
                                     ',initarg
                                     ',slot-spec
                                     ',sc-list))
#+*initarg-as-keywords*
               ,@keywords
               (rem-format ',name)
               ;; define generic functions for slots access.
               ,@slots
               ',name)))

(defun %compute-slot-declarations (sc-list slot-spec)
   ;; Compute slot descriptions (deal with multiple inheritance)
   ;; NOTE: this function is too complex and probably wrong.
   (let ((supers  (%compute-super-classes  sc-list))
         (direct  (%compute-direct-classes sc-list))
         (gfnames ())
         (slots   ())
         (sslots  ()))
        ;; Build  slot  declarations  for  direct inherited slots,  no
        ;; accessors here.
        (dolist (x direct)
                (let ((inherited (%compute-class-descriptions x)))
                     ;; Remove common base classes, it should not appends.
                     (setq supers (remove x supers :test #'equal))
                     ;; Compute unmodified inherited slots (with gf accessors).
                     (setq sslots (append sslots inherited))
                     ;; Compute inherited slots (without gf accessor).
                     (setq slots
                           (append (mapcar #'%compute-direct-slot-declarations
                                           inherited)
                                   slots))))
        ;; Add slots that belongs to the class
        (setq slots  (append slots slot-spec))
        ;; compute gf accessors either declared or inherited.
        (setq sslots (append sslots slot-spec))
        (dolist (x sslots)
                (setq gfnames
                      (append gfnames (%compute-generic-functions x nil))))
        ;; Add slots from multiple inheritance (if any) and remove
        ;; gf conflitcs from declared or inherited classes.
        (dolist (x supers)
                (let ((new-slots (%compute-class-descriptions x)))
                     (setq new-slots
                           (mapcar (lambda (s)
                                      (%compute-slot-no-conflitcs s gfnames))
                                   new-slots))
                     (setq slots (append slots new-slots))))
        slots))

(defun %compute-direct-slot-declarations (slot)
   ;; Remove  generic  function  declarations  inherited  form  direct
   ;; ancestors.  That  is,  build  a  new  slot  declaration  without
   ;; :accessor :reader :writer or :boundp options.
   (cond
         ((null slot)
          ())
         ((member (car slot) *slot-option-keywords*)
          (%compute-direct-slot-declarations (cddr slot)))
         (t (cons (car slot) (%compute-direct-slot-declarations (cdr slot))))))

(defun %compute-generic-functions (slot gf)
   ;; Compute generic  function  declarations  inherited  form  direct
   ;; ancestors.
   (cond
         ((null slot)
          gf)
         ((member (car slot) *slot-option-keywords*)
          (%compute-generic-functions (cddr slot) (cons (cadr slot) gf)))
         (t (%compute-generic-functions (cdr slot) gf))))

(defun %compute-slot-no-conflitcs (slot gf)
   ;; compute slot decription removing gf accessors already defined.
   (cond
         ((null slot)
          ())
         ((and (member (car slot) *slot-option-keywords*)
               (member (cadr slot) gf))
          (%compute-slot-no-conflitcs (cddr slot) gf))
         (t (cons (car slot) (%compute-slot-no-conflitcs (cdr slot) gf)))))

(defun %compute-direct-classes (sc-list)
   ;; Given  a super class list 'sc-list',  returns the list of direct
   ;; super classes.
   (cond
         ((symbolp sc-list)
          sc-list)
         (t
            (setq sc-list (car sc-list))
            (cons (%compute-direct-classes sc-list)
                  (%compute-direct-classes (%compute-class-super sc-list))))))

(defun %compute-super-classes (sc-list)
   ;; Given  a  super  class  list 'sc-list',  returns the list of all
   ;; inherited slots.
   (cond
         ((null sc-list) ())
         ((symbolp sc-list)
          (cons (%compute-direct-classes sc-list)
                (%compute-super-classes (%compute-class-super sc-list))))
         (t (append (%compute-super-classes (car sc-list))
                    (%compute-super-classes (cdr sc-list))))))

(defun %compute-slot-options (name slot index)
   ;; Compute the slot options.
   ;; Valid keywords are :
   ;;
   ;;   :ACCESSOR accessor-name
   ;;   :INITARG  symbol
   ;;   :INITFORM form
   ;;   :READER   reader-name
   ;;   :WRITER   writer-name
   ;;   :BOUNDP   boundp-name
   ;;
   (let ((accessors ())
         (initform  ''|#<unbound-value>|)
         (initarg   ())
         (key       ())
         (slot-name ())
         (val       ()))
        (cond
              ((consp slot)
               ;; skip slot name
               (setq slot-name (car slot))
               (setq slot      (cdr slot)))
              (t
               (error "defclass: invalid slot argument list ~a~%" slot)))
        ;; parse slot options
        (while (consp slot)
               (unless (consp (cdr slot))
                       (error "defclass: invalid slot argument list ~a~%" slot))
               (setq key  (car  slot))
               (setq val  (cadr slot))
               (setq slot (cddr slot))
               (case key
                     ((:initarg)  (setq initarg val))
                     ((:initform) (setq initform (%compute-initform val)))
                     ((:boundp)   (push `(defmethod ,val
                                            ((obj ,name))
                                            (slot-boundp obj ',slot-name))
                                        accessors))
                     ((:reader)   (push `(defmethod ,val
                                            ((obj ,name))
                                            (get-slot-index obj ,index))
                                         accessors))
                     ((:writer)   (push `(defmethod ,val
                                            ((val <object>) (obj ,name))
                                            (set-slot-index val obj ,index))
                                         accessors))
                     ((:accessor) (push `(defmethod ,val
                                            ((obj ,name))
                                            (get-slot-index obj ,index))
                                         accessors)
                                  (push `(defmethod ,(concat '%setf- val)
                                            ((obj ,name) (val <object>))
                                            (set-slot-index val obj ,index))
                                         accessors)
                                  (push `(defsetf ,val ,(concat '%setf- val))
                                        accessors))
                     (t (error "defclass: invalid slot option ~a~%" key))))
        (list initform
              initarg
              (append
                      (%compute-missing-generic-functions accessors)
                      accessors))))

(defun %assure-defgeneric (name arg)
  `(if (not (and (fboundp ',name)
                 (generic-function-p (symbol-function ',name))))
       (defgeneric ,name ,arg)))

(defun %compute-missing-generic-functions (accessors)
   ;; Automatically create generic functions for accessors.
   (let ((l ()))
        (dolist (x accessors)
                (if (eq (car x) 'defmethod)
                    (let ((y (cadr x)))
                         (case (length (caddr x))
                               ((1) (push (%assure-defgeneric y '(x)) l))
                               ((2) (push (%assure-defgeneric y '(x y)) l))
                               (t   (error "defclass: syntax error ~s." x))))))
        l))

;;; Interface to class information.

(defun %compiling ()
   ;; check if we are compiling or not.
   (and (boundp '*compile-file*) (dynamic *compile-file*)))

(defun %class-info (name)
   ;; returns class vector information.
   (if (%compiling)
       (or (gethash name *compiler-struct-ht*) (class-info (find-class name)))
       (class-info (find-class name))))

(defun %compute-class-slots (name)
   ;; returns slot names for class 'name'.
   (elt (%class-info name) 1)) ;; class-info-slots

(defun %compute-class-initforms (name)
   ;; returns slot initforms for class 'name'.
   (elt (%class-info name) 2)) ;; class-info-initforms

(defun %compute-class-descriptions (name)
   ;; returns info description for class 'name'.
   (elt (%class-info name) 4)) ;; class-info-descriptions

(defun %compute-class-super (name)
   ;; returns super for class 'name'.
   (elt (%class-info name) 5)) ;; class-info-super

(defun %build-class (name sc-list abstractp info)
   ;; create a new class object for 'name'.
   (cond
         ((%compiling)
          (setf (gethash name *compiler-struct-ht*) info))
         (t
          ;; add a new class to the object system.
          (create-class name sc-list abstractp)
          ;; define the class-info for this class.
          (setf (class-info (find-class name)) info)))
   t)

;;; Common slot initforms.

(defun %zero-initform  () 0)
(defun %one-initform   () 1)
(defun %false-initform () nil)
(defun %true-initform  () t)

(defun %compute-initform (initform)
   ;; Returns  the  appropriate  init  function  text  for  the  given
   ;; initform. Does some minimal code sharing for very common cases.
   (cond
         ((eq initform nil) '(function %false-initform))
         ((eq initform t)   '(function %true-initform))
         ((eq initform 0)   '(function %zero-initform))
         ((eq initform 1)   '(function %one-initform))
         ((and (consp initform) (null (cdr initform)))
          ;; It's a no argument function call: (xxx)
          `(function ,(car initform)))
         (t `(lambda () ,initform))))

;;; Add feature

(provide 'defclass)
