;;;; -*-Mode:LISP; Package:LISP; Base:10; Syntax:ISLISP -*-
;;;; Date:      2006/10/29
;;;; Title:     defstruct.lsp
;;;; Author:    C. Jullien
;;;; RCS:       "$Id: defstruc.lsp,v 1.6 2006/10/30 08:02:31 cjullien Exp $"

;;; DEFSTRUCT package.

;;; When  *structure-standard-names*  is T,  CommonLisp convention is
;;; used.  In  that  case,  (defstruct  foo x y) will generate foo-x,
;;; foo-y,  foo-p,  copy-foo  and  make-foo  functions  and :x and :y
;;; keywords.  Otherwise,  we  generate  Le-Lisp  functions  #:foo:x,
;;; #:foo:y  and  #:foo:make.  Note  that  this DEFSTRUCT package has
;;; inheritance only for CLtL syntax.

(require 'setf)

(in-package "openlisp")

(export '(defstruct structurep make *structure-standard-names* %build-class )
        "openlisp")

(defpackage "defstruct"
   (:use    "openlisp")
;   (:export defstruct structurep *structure-standard-names* %build-class)
   (:import-from "openlisp"
            *compile-file* *compiler-struct-ht* add-format))

#+fii

(in-package "defstruct")

;;; Define naming convention. Value () is for Le-Lisp and t for CLtL.

(defdynamic *structure-standard-names* t)

;;; The DEFSTRUCT macro.

(defmacro openlisp:defstruct (name &rest slots)
   (%expand-defstruct name slots))

;;; User interface

(defmacro openlisp:structurep (struct)
   ;; the structure predicate
  `(instancep ,struct <standard-structure>))

(defun openlisp:make (name &rest keys)
   (initialize-object (allocate-object (find-class name)) keys))

;;; Helper functions

(defun %expand-defstruct (name slots)
   ;; define a new structure
   (let ((initform  (mapcar #'%compute-slot-init-value slots))
         (index     -1)
         (options   ())
         (std-name  ())
         (predicate '%use-default-name)
         (conc-name '%use-default-name)
         (make      '%use-default-name)
         (copy      '%use-default-name)
         (initarg   ())
         (struct    ())
         (printer   ())
         (boa-make  ())
         (super     '<standard-structure>)
         (slot      ()))
        (when (consp name)
              ;; CLtL   syntax,   check   for   :include,  :predicate,
              ;; :constuctor, :print-function, :copier and :conc-name.
              ;; Ignore other options.
              (setq options (cdr name))
              (setq name    (car name))
              (if (not (symbolp name))
                  (error "defstruct: structure name is not a symbol ~s ~s.~%"
                         name
                         (class-of name)))
              (for ((option options (cdr option)))
                   ((null option))
                   (if (not (consp (car option)))
                       (error "defstruct: invalid option : ~a~%." (car option)))
                   (let ((key (caar option))
                         (val (cadar option)))
                        (case key
                              ((:INCLUDE)
                               (if (not (symbolp val))
                                   (error "defstruct: not a symbol ~a~%" val))
                               (setq super val)
                               (setq initform
                                     (append
                                        (%update-inherited-initforms
                                                (%compute-class-slots     val)
                                                (%compute-class-initforms val)
                                                (cddar option))
                                        initform))
                               (setq slots
                                     (append (%compute-class-slots val) slots)))
                              ((:CONSTRUCTOR)
                               (setq make val)
                               ;; boa-constructor test.
                               (if (and (consp (cddar option))
                                        (consp (caddar option)))
                                   (setq boa-make
                                         (caddar option))))
                              ((:PREDICATE)
                               (setq predicate val))
                              ((:CONC-NAME)
                               (setq conc-name val))
                              ((:COPIER)
                               (setq copy      val))
                              ((:PRINT-FUNCTION)
                               (setq printer   val))
                              (t
                               (error "defstruct: invalid option ~a~%" key))))))
        ;; undecorated structure name.
        (setq std-name (%normalize-name name))
        (if (eq predicate '%use-default-name)
            (setq predicate '-p))
        (if (eq conc-name '%use-default-name)
            (setq conc-name (concat std-name "-")))
        (if (eq make '%use-default-name)
            (setq make 'make-))
        (if (eq copy '%use-default-name)
            (setq copy 'copy-))
        (setq slot (mapcar (lambda (x)
                             (let ((read-only ()))
                                  (when (consp x)
                                        (setq read-only (member :read-only x))
                                        (when read-only
                                              (setq read-only
                                                    (cadr read-only)))
                                        (setq x (car x)))
                                  (setq index  (1+ index))
                                  (setq struct (cons x struct))
                                  (if read-only
                                      (%make-slot-reader
                                            std-name conc-name x index)
                                      (%make-slot-accessor
                                            std-name conc-name x index))))
                           slots))
        (setq initarg (mapcar (lambda (x)
                                 ;; generate keyword for each field.
                                 (define-keyword (if (consp x) (car x) x)))
                               slots))
        (when super
              ;; inherit missing behavior from super
              (unless printer
                      (setf printer (get-format super))))
        (when printer
              (setq printer
                    `(add-format ',name ,(if (consp printer)
                                             ;; assumes lambda
                                             printer
                                             (list 'quote printer)))))
        (setq struct (nreverse struct))
       `(progn
               (%build-class ',name
                             '(,super)
                             nil
                             (create-class-info
                                    ',struct
                                    ',initform
                                    ',initarg
                                    nil
                                    nil))
               ,(if make
                   (if boa-make
                       `(defun ,(%generate-symbol std-name make :constructor)
                          (&rest key)
                          (boa-constructor (allocate-object (class ,name)) key))
                       `(defun ,(%generate-symbol std-name make :constructor)
                          (&rest key)
                          (initialize-object (allocate-object (class ,name))
                                             key))))
               ,(if copy
                  `(defun ,(%generate-symbol std-name copy :copier) (obj)
                     (if (eq (class-of obj) ',name)
                         (copy-seq obj)
                         (error "~A is not a '~A' object~%" obj ',name))))
               ,(if predicate
                  `(defun ,(%generate-symbol std-name predicate :predicate) (x)
                     (instancep x (class ,name))))
               (rem-format ',name)
               ,printer
               ,@slot
               ',name)))

(defun %normalize-name (sym)
   ;; strip #\< and #\> around the name: <foo> -> foo
   (let ((str (convert sym <string>)))
        (if (eql (elt str 0) #\<)
            (intern (string-trim "<>" str))
            sym)))

(defun %generate-symbol (struct conc-name name)
   (if (dynamic *structure-standard-names*)
       ;; CommonLisp name is in use -> ship-x
       (case name
             ((:CONSTRUCTOR)
               (if (eq conc-name 'make-)
                   (concat conc-name struct)
                   conc-name))
             ((:COPIER)
               (if (eq conc-name 'copy-)
                   (concat conc-name struct)
                   conc-name))
             ((:PREDICATE)
               (if (eq conc-name '-p)
                   (concat struct conc-name)
                   conc-name))
             (t
               (if conc-name
                   ;; generate a prefixed symbol (ship-x)
                   (concat conc-name name)
                   ;; no prefix must be used.
                   name)))
       ;; Le-Lisp package is in use -> #:ship:x
       (case name
             ((:CONSTRUCTOR) (intern 'make struct))
             (t (intern name struct)))))

(defun %compute-slot-init-value (slot)
   ;; ignore :read-only and :type options
   (cond ((symbolp slot)
          ())
         ((or (atom slot) (null (symbolp (car slot))))
          (error "defstruct: not an atom ~A~%" slot))
         (t
          (when (consp (cdr slot))
                (cadr slot)))))

(defun %update-inherited-initforms (slots initform option)
   (setq initform (copy-seq initform))
   (dolist (x option)
           (if (and (consp x) (consp (cdr x)))
               (let ((pos (position (car x) slots)))
                    (if pos
                        (set-elt (cadr x) initform pos)
                        (error "defstruct: ~a is not an inherited slot.~%"
                               (car x))))
               (error "defstruct: invalid inherited option :~a.~%" x)))
   initform)

(defglobal %use-defslot t)      ;; t if you use DEFSLOT OpenLisp extension.

(defun %make-slot-accessor (struct conc-name slot index)
   ;; create  the  access  function  for  a  new slot.
   (let* ((get-slot (%generate-symbol struct conc-name slot))
          (set-slot (concat 'set- get-slot)))
        (if %use-defslot
           `(defslot ,get-slot :accessor ,index)
           `(progn
                   (defun ,get-slot (obj) (svref obj ,index))
                   (defun ,set-slot (obj val) (svset val obj ,index))
                   (defsetf ,get-slot ,set-slot)))))

(defun %make-slot-reader (struct conc-name slot index)
   ;; create  the  reader  function  for  a  new slot.
   (let ((get-slot (%generate-symbol struct conc-name slot)))
        (if %use-defslot
           `(defslot ,get-slot :reader ,index)
           `(defun ,get-slot (obj) (svref obj ,index)))))

;;; Interface to class information.

(defun %compiling ()
   ;; check if we are compiling or not.
   (and (boundp '*compile-file*) (dynamic *compile-file*)))

(defun %class-info (name)
   ;; returns class vector information.
   (if (%compiling)
       (or (gethash name *compiler-struct-ht*) (class-info (find-class name)))
       (class-info (find-class name))))

(defun %compute-class-slots (name)
   ;; returns slot names for class 'name'.
   (elt (%class-info name) 1)) ;; class-info-slots

(defun %compute-class-initforms (name)
   ;; returns slot initforms for class 'name'.
   (elt (%class-info name) 2)) ;; class-info-initforms

(defun %compute-class-descriptions (name)
   ;; returns info description for class 'name'.
   (elt (%class-info name) 4)) ;; class-info-descriptions

(defun %compute-class-super (name)
   ;; returns super for class 'name'.
   (elt (%class-info name) 5)) ;; class-info-super

(defun %build-class (name sc-list abstractp info)
   ;; create a new class object for 'name'.
   (cond
         ((%compiling)
          (setf (gethash name *compiler-struct-ht*) info))
         (t
          ;; add a new class to the object system.
          (create-class name sc-list abstractp)
          ;; define the class-info for this class.
          (setf (class-info (find-class name)) info)))
   t)

;;; Add defstruct feature.

(provide 'defstruc)
