;;;; -*-Mode:LISP; Package:LISP; Base:10; Syntax:ISLISP -*-
;;;; Date:      2006/04/11
;;;; Title:     flet.lsp
;;;; Author:    C. Jullien
;;;; RCS:       "$Id: flet.lsp,v 1.3 2006/09/02 07:17:41 jullien Exp $"

;;;
;;; FLET / LABELS Special Forms (written as MACRO).
;;;
;;; This  definition is only used by eval.  The compiler uses its own
;;; and more accurate definition.  You can notice differences between
;;; the  two  versions.  The  compiled  one  should  be  taken as the
;;; reference.
;;;

(require 'setf)

(in-package "openlisp")

(defpackage "flet"
   (:use    "openlisp")
   (:export flet labels))

(in-package "flet")

(defmacro flet (lfun &rest body)
  ;; flet Special Form.
  `(let ,(mapcar (lambda (x) `(,(%fl-concat-f (car x)) (lambda ,@(cdr x))))
                 lfun)
        ,@(%fl-body body (mapcar #'car lfun))))

(defmacro labels (lfun &rest body)
  ;; labels Special Form.
  (let* ((fs  (mapcar #'car lfun))
         (cfs (mapcar #'%fl-concat-f fs)))
        `(let ,cfs
              ,@(mapcar (lambda (n x)
                          `(setq ,n ,(%fl-one-form `(lambda ,@(cdr x)) fs)))
                        cfs
                        lfun)
              ,@(%fl-body body fs))))

(defun %fl-concat-f (x)
   ;; generates lexical function name.
   (if (symbolp x)
       (concat '%fl-function- x)
       (error "Invalid identifier ~a" x)))

(defun %fl-body (form lfun)
   ;; expands a list of forms using lexical function list lfun.
   (mapcar (lambda (x) (%fl-one-form x lfun)) form))

(defun %fl-expand-rest (form lfun)
   ;; generates  a  binding  form  where initialization value is done
   ;; using  lexical  function  list  lfun.  The  first form is never
   ;; expanded. It's used by binding forms such as LET, LET*, DO ...
   (mapcar (lambda (x)
              (if (atom x)
                  x
                  (cons (car x) (%fl-body (cdr x) lfun))))
           form))

(defun %fl-one-form (f l)
   ;; expands one form using lexical function defined in list l.
   (let ((op (if (consp f) (car f))))
        (cond
           ((atom f)
            ;; standard variable, lookup in lexical environment
            f)
           ((consp op)
            (%fl-body f l))
           ((member op l)
            ;; a  call  to  a  lexically bound function.  Use funcall
            ;; with its associated function.
            `(funcall ,(%fl-concat-f op) ,@(%fl-body (cdr f) l)))
           ((and (symbolp op) (macro-function op))
            ;; expand macro and start again.
            (%fl-one-form (macroexpand f) l))
           (t
            (case op
               ((CLASS DYNAMIC GO QUOTE)
                ;; variables   inside   those  forms  should  not  be
                ;; converted.
                f)
               ((LAMBDA)
                `(,op ,(cadr f) ,@(%fl-body (cddr f) l)))
               ((FUNCTION)
                (cond
                      ((consp (cadr f))
                       ;; a lambda, expand its body.
                       `(function (%fl-one-form (cadr f) l)))
                      ((member (cadr f) l)
                       ;; a lexically definied function.
                       (%fl-concat-f (cadr f)))
                      (t
                       ;; a globally defined symbol.
                       f)))
               ((CASE CASE-USING)
                ;; expand  the value and the body forms.  Case values
                ;; are leave unmodified.
                `(,op ,(%fl-one-form (cadr f) l)
                      ,@(%fl-expand-rest (cddr f) l)))
               ((LET LET* DYNAMIC-LET)
                ;; expand initialization values and body.
                `(,op ,(%fl-expand-rest (cadr f) l)
                      ,@(%fl-body (cddr f) l)))
               ((DO DO* FOR)
                ;; TO BE CHECKED!!! expand initialization values and body.
                `(,op ,(%fl-expand-rest (cadr f) l)
                      ,@(%fl-body (cddr f) l)))
               ((WITH-OPEN-INPUT-FILE WITH-OPEN-OUTPUT-FILE WITH-OPEN-IO-FILE)
                ;; stream variable should not be converted.
                `(,op ,(car (%fl-expand-rest (list (cadr f)) l))
                      ,@(%fl-body (cddr f) l)))
               (t
                ;; no  special  case,  leave  the function unmodified
                ;; and expand the body.
                `(,op ,@(%fl-body (cdr f) l))))))))

(provide 'flet)
