;;;; -*-Mode:LISP; Package:LISP; Base:10; Syntax:ISLISP -*-
;;;; Date:      2006/04/11
;;;; Title:     pretty.lsp
;;;; Author:    C. Jullien
;;;; RCS:       "$Id: pretty.lsp,v 1.5 2006/12/31 12:22:40 jullien Exp $"

;;; Trivial pretty-printer

(require 'setf)

(in-package "openlisp")

(defpackage "pretty"
   (:use    "openlisp")
   (:export pretty pprint))

(in-package "pretty")

(defglobal *max-pos*  79)
(defglobal *pos*       0)
(defglobal *ppstream* ())

(set-property 2 'lambda         'special)
(set-property 2 'tagbody        'special)
(set-property 4 'case-using     'special)
(set-property 6 'unwind-protect 'special)
(set-property 6 'return-from    'special)
(set-property 7 'dynamic-let    'special)

;; Main functions

(defun pretty (object &rest stream)
   ;; pretty print a function or any list expression.
   (when (symbolp object)
         (let ((def (function-definition object)))
              (when (consp def)
                    (pprint def (if stream (car stream) (standard-output))))))
   object)

(defun pprint (x &rest stream)
   ;; pretty print an expression.
   (setq *ppstream* (if stream (car stream) (standard-output)))
   (setq *pos*      0)
   (%pp-object x ())
   (%pp-newline)
   (%pp-newline)
   t)

;; Implementation

(defun %pp-atom (s)
   ;; pretty print an atom.
   (cond
         ((characterp s)
          (format *ppstream* "~C" s)
          (incf *pos*))
         ((stringp s)
          (%pp-string s t))
         ((symbolp s)
          (setf s (format () "~S" s))
          (%pp-string s ()))
         ((vectorp s)
          (format *ppstream* "~S" s)
          (incf *pos* (%pp-flat-size s)))
         (t
          (format *ppstream* "~A" s)
          (incf *pos* (%pp-flat-size s)))))

(defun %pp-string (s flag)
   ;; pretty print a strings.
   (format *ppstream* (if flag "~S" "~A") s)
   (incf *pos* (length s)))

(defun %pp-flat-size (s)
   ;; computes the length of an expression.
   (let ((str (create-string-output-stream))
         (res ()))
        (format str "~A" s)
        (setq res (get-output-stream-string str))
        (close str)
        (length res)))

(defglobal %pp-chars "@abcdefghijklmnopqrstuvwxyz")

(defun %pp-object (x cst)
   ;; pretty print a single object
   (cond
         ((null x)
          (%pp-atom #\()
          (%pp-atom #\)))
         ((characterp x)
          (case x
                ((#\newline)         (%pp-string "#\\newline"         ()))
                ((#\space)           (%pp-string "#\\space"           ()))
                ((#\tab)             (%pp-string "#\\tab"             ()))
                ((#\carriage-return) (%pp-string "#\\carriage-return" ()))
                ((#\escape)          (%pp-string "#\\escape"          ()))
                ((#\rubout)          (%pp-string "#\\rubout"          ()))
                ((#\backspace)       (%pp-string "#\\backspace"       ()))
                ((#\control-\)       (%pp-string "#\\control-\\"      ()))
                ((#\control-])       (%pp-string "#\\control-]"       ()))
                ((#\control-^)       (%pp-string "#\\control-^"       ()))
                ((#\control-_)       (%pp-string "#\\control-_"       ()))
                (t                   (%pp-string "#\\" ())
                                     (let ((val (char-int x)))
                                          (cond
                                                ((<= val 26)
                                                 (%pp-string "control-" ())
                                                 (%pp-atom (elt %pp-chars val)))
                                                ((< val 32)
                                                 (setf val
                                                       (int-char (+ val 64)))
                                                 (%pp-string "control-" ())
                                                 (%pp-atom val))
                                                (t
                                                 (%pp-atom x)))))))
         ((consp x)
          (case (car x)
                ((QUOTE)
                 (%pp-atom #\')
                 (%pp-object (cadr x) t))
                ((FUNCTION)
                 (if (consp (cdr x))
                     (progn
                            (%pp-atom #\#)
                            (%pp-atom #\')
                            (%pp-object (cadr x) cst))
                     (if (> (%pp-flat-size x) (- *max-pos* *pos*))
                         ;; does not fit in a single line
                         (%pp-list x cst t)
                         (%pp-list x cst ()))))
                (t
                 (if (> (%pp-flat-size x) (- *max-pos* *pos*))
                     ;; does not fit in a single line
                     (%pp-list x cst t)
                     (%pp-list x cst ())))))
          (t
           (%pp-atom x))))
   
(defun %pp-list (l cst vertical-p)
   ;; pretty print a list of objects. Take care of Special forms.
   (%pp-atom #\( )
   (let ((pos1 *pos*))
        (%pp-object (car l) cst)
        (if (and vertical-p (consp (car l)) (cdr l))
            (%pp-indent (+ pos1 1)))
        (let ((pos2 (+ *pos* 1))
              (key  (car l))
              (tag  ())
              (flag ()))
             (if cst
                 (setq l (cdr l))
                 (case key
                       ((DEFUN DEFMACRO DEFGENERIC DEFMETHOD)
                        (if (and (consp (cdr l)) (consp (cddr l)))
                            (progn
                                   (%pp-atom #\space)
                                   (%pp-object (cadr l) cst)
                                   (%pp-atom #\space)
                                   (%pp-arglist (caddr l) cst)
                                   (setq l (cdddr l))
                                   (setq pos2 3)
                                   (%pp-indent (- pos2 1)))
                            (setq l (cdr l))))
                       ((TAGBODY)
                        (%pp-indent (+ pos1 (property key 'special) -1))
                        (setq vertical-p t)
                        (setq l (cdr l))
                        (setq tag t))
                       ((CASE)
                        (%pp-atom #\space)
                        (%pp-object (cadr l) cst)
                        (%pp-indent (- pos2 1))
                        (setq vertical-p t)
                        (setq l (cddr l)))
                       ((CASE-USING)
                        (%pp-atom #\space)
                        (%pp-object (cadr l) cst)
                        (%pp-atom #\space)
                        (%pp-object (caddr l) cst)
                        (%pp-indent  (+ pos1 (property key 'special) -1))
                        (setq vertical-p t)
                        (setq l (cdddr l)))
                       ((LET LET* FOR DO DO*)
                        (%pp-atom #\space)
                        (cond
                              ((consp (cadr l))
                               (%pp-atom #\()
                               (let ((lst (cadr l)))
                                    (while (consp lst)
                                           (%pp-object (car lst) cst)
                                           (and (setq lst (cdr lst))
                                                (%pp-indent (+ pos2 1)))))
                               (%pp-atom #\)))
                              (t
                               (%pp-object (cadr l) cst)))
                        (%pp-indent (- pos2 1))
                        (setq vertical-p t)
                        (setq l (cddr l)))
                       (t
                        (setq l (cdr l)))))
             (while (consp l)
                    (when flag
                          (%pp-indent (if (and (symbolp key)
                                               (property key 'special))
                                          (+ pos1 (property key 'special) -1)
                                          (- pos2 1))))
                    (when (and tag (consp (car l)))
                          (%pp-spaces 4))
                    (%pp-atom #\space)
                    (%pp-object (car l) cst)
                    (setq flag vertical-p)
                    (setq l (cdr l)))
             (when l
                   ;; dotted pair
                   (%pp-atom #\space)
                   (%pp-atom #\.)
                   (%pp-atom #\space)
                   (if flag
                       (%pp-indent pos2))
                   (%pp-object l cst))
             (%pp-atom #\) ))))

(defun %pp-arglist (args cst)
   ;; print function argument list (adding &rest when needed)
   (if (not (listp args))
       (%pp-object (list '&rest args) cst)
       (let ((b (copy-seq args))
             (d ())
             (l ()))
            (setq l b)
            (while (consp l)
                   (setq d l)
                   (setq l (cdr l)))
            (when l
                  (setf (cdr d) (list '&rest l)))
            (%pp-object b cst))))

(defun %pp-indent (x)
   ;; intent up to position x.
   (%pp-newline)
   (%pp-spaces x))

(defun %pp-spaces (x)
   ;; add x spaces.
   (for ((i 0 (+ i 1)))
        ((>= i x))
        (%pp-atom #\space)))

(defun %pp-newline ()
   ;; insert a newline.
   (format *ppstream* "~%")
   (setq *pos* 0))

(provide 'pretty)
