;;;; -*-Mode:LISP; Package:LISP; Base:10; Syntax:ISLISP -*-
;;;; Date:      2006/11/08
;;;; Title:     startup.lsp
;;;; Author:    C. Jullien
;;;; RCS:       "$Id: startup.lsp,v 1.15 2007/01/01 10:11:01 jullien Exp $"

;;;
;;; Startup for standard environnment.
;;;

(in-package "openlisp")

;;;
;;; Define  backconstant as the very first function in startup.lsp so
;;; that we can use its function type to know if this file is running
;;; in compiled or in interpreted mode (see *compiled-env* below).
;;;

(defun backconstant (s)
   ;; t if expression is like (QUOTE xxx)
   (and (consp s) (eq (car s) 'quote) (null (cddr s))))

(defglobal *compiled-env*
   (if (member (function-type 'backconstant) '(subr1 subrn))
       t))

(if (and (not *quiet-mode*) (not *compiled-env*))
    (let ((info (machine-info)))
         (format (standard-output) ";; Copyright (c) 1988-2006.~%")
         (format (standard-output) ";; System '~a' (~d-bit) on '~a', ~a.~%"
                 (system-name)
                 (* (pointer-size) 8)
                 (elt info 3) ;; machine name
                 (if (= (character-size) 1) "ASCII" "UNICODE"))
         (format (standard-output) ";; Reading startup ..~%")))

;;;
;;; ISLISP Level : ISO/IEC 13816:1997(E) ISLISP Programing Language.
;;;

(defconstant *islisp-version* 199705) ;; 1st May 1997

;;;
;;; Installation directory
;;;

(defdynamic *system-directory*
   ;; can be something like "/usr/local/openlisp"
   (if (boundp '*system-directory*)
       (dynamic *system-directory*)
       (or (getenv "OPENLISP") ".")))

(defdynamic *system-path*
   ;; list of default directories.
   (cons "."
         (mapcar (lambda (x)
                    (string-append (dynamic *system-directory*) x))
                 '("/lib" "/net" "/contrib" "/bench" "/tst"))))

;;;
;;; We need backquote macro early in bootstrap process.
;;;

(defun backquotify (s)
   ;; backquotify an expression.
   (cond
         ((null s)
          ())
         ((vectorp s)
          ;; ISLISP section 8: ` or quasiquote constructs a list structure.
          ;; This case is CLtL compatible OpenLisp extension.
          (list 'apply #'vector (backquotify (convert s <list>))))
         ((atom s)
          (list 'quote s))
         ((eq (car s) '|,|)
          (cdr s))
         ((or (eq (car s) '|,@|) (eq (car s) '|,.|))
          (error "` syntax error: ~s~%" s))
         ((and (consp (car s)) (eq (caar s) '|,@|))
          (let ((a (cdar s)) (d (backquotify (cdr s))))
               (if d
                   (list 'append a d)
                   a)))
         ((and (consp (car s)) (eq (caar s) '|,.|))
          (if (cdr s)
              (list 'nconc (cdar s) (backquotify (cdr s)))
              (cdar s)))
         (t
          (let ((a (backquotify (car s)))
                (d (backquotify (cdr s))))
               (cond
                     ((null d)
                      (if (backconstant a)
                          (list 'quote (list (cadr a))) ; (cons 'a ()) -> '(a)
                          (list 'list a))) ; (cons a ())  -> (list a)
                     ((and (backconstant a) (backconstant d))
                      ;; (cons 'a 'd) -> '(a . d)
                      (list 'quote (cons (cadr a) (cadr d))))
                     ((and (consp d) (eq (car d) 'list))
                      ;; (cons a (list d1 .. dN)) -> (list a d1 .. dN)
                      (cons 'list (cons a (cdr d))))
                     (t (list 'cons a d)))))))

(defun backquote-dispatcher (stream char)
   (case char
         ((#\`)
          (backquotify (read stream)))
         ((#\@)
          (list '|@| (read stream)))
         ((#\,)
          (case (preview-char stream)
              ((#\@) (read-char stream)
                     (cons '|,@| (read stream)))
              ((#\.) (read-char stream)
                     (cons '|,.| (read stream)))
              (t     (cons '|,|  (read stream)))))
         (t
          (error "backquote-dispatcher: unsupported character ~s." char))))

(set-macro-character #\` #'backquote-dispatcher)
(set-macro-character #\, #'backquote-dispatcher)
(set-macro-character #\@ #'backquote-dispatcher)

(defun macroexpand-1 (x)
   (if (and (consp x) (symbolp (car x)) (macro-function (car x)))
       (funcall (macro-function (car x)) x)
       x))

(defun macroexpand (x)
   (while (and (consp x) (symbolp (car x)) (macro-function (car x)))
          (setq x (funcall (macro-function (car x)) x)))
   x)

(defun macroexpand-all (x)
   (cond
         ((atom x)
          x)
         ((and (symbolp (car x)) (macro-function (car x)))
          (macroexpand-all (funcall (macro-function (car x)) x)))
         ((and (symbolp (car x)) (member (car x) '(case case-using)))
          (cons (car x)
                (cons (macroexpand-all (cadr x))
                      (mapcar (lambda (l)
                                 (cons (car l) (macroexpand-all (cdr l))))
                              (cddr x)))))
         (t
          (let ((l ()))
               (while (consp x)
                      (setq l (cons (macroexpand-all (car x)) l))
                      (setq x (cdr x)))
               (prog1 (nreverse l)
                      (set-cdr x l))))))

;;;
;;; The eval-when Special Form written as macro
;;;

(defmacro eval-when (situations &rest body)
   ;; conditionally execute code form a situation list.
   ;; Valid situation are COMPILE EVAL LOAD
   (if (member (dynamic *situation*) situations)
      `(progn ,@body)))

;;;
;;; Streams macros
;;;

(defmacro with-standard-input (stream &rest body)
   `(dynamic-let ((*standard-input* ,stream)) ,@body))

(defmacro with-standard-output (stream &rest body)
   `(dynamic-let ((*standard-output* ,stream)) ,@body))

(defmacro with-error-output (stream &rest body)
   `(dynamic-let ((*error-output* ,stream)) ,@body))

(defmacro with-open-output-file (stream &rest body)
   `(let ((,(car stream) (open-output-file ,@(cdr stream))))
         (unwind-protect (progn ,@body)
                         (close ,(car stream)))))

(defmacro with-open-input-file (stream &rest body)
   `(let ((,(car stream) (open-input-file ,@(cdr stream))))
         (unwind-protect (progn ,@body)
                         (close ,(car stream)))))

(defmacro with-open-io-file (stream &rest body)
   `(let ((,(car stream) (open-io-file ,@(cdr stream))))
         (unwind-protect (progn ,@body)
                         (close ,(car stream)))))

(defmacro with-output-to-string (stream &rest body)
   `(let ((,(car stream) (create-string-output-stream)))
         (unwind-protect (progn ,@body)
                         (close ,(car stream)))))

(defmacro with-input-from-string (stream &rest body)
   `(let ((,(car stream) (create-string-input-stream ,(cadr stream))))
         (unwind-protect (progn ,@body)
                         (close ,(car stream)))))

(defmacro with-open-input-pipe (stream &rest body)
   `(let ((,(car stream) (open-input-pipe ,@(cdr stream))))
         (unwind-protect (progn ,@body)
                         (close ,(car stream)))))

(defmacro with-open-output-pipe (stream &rest body)
   `(let ((,(car stream) (open-output-pipe ,@(cdr stream))))
         (unwind-protect (progn ,@body)
                         (close ,(car stream)))))

;;;
;;; Define generic functions for CREATE and INITIALIZE-OBJECT
;;;

(defgeneric create (cl &rest init)
   ;; Create is used by both 'defstruct' and 'defclass' define forms.
   (:method ((cl <object>) &rest init)
            (identity init)
            (error "CREATE : not a <standard-class> object ~a~%" cl))
   (:method ((cl <standard-class>) &rest init)
            (let ((obj (allocate-object cl)))
                 (initialize-object obj init)
                 obj))
   ;; Extension to ISLISP
   (:method ((cl <symbol>) &rest init)
            (let ((obj (allocate-object (find-class cl))))
                 (initialize-object obj init)
                 obj)))

(defgeneric initialize-object (instance initvalues)
   (:method ((instance <object>) initvalues)
            (default-initialize-object instance initvalues)))

;;;
;;; Control structures (extensions).
;;;

(defmacro when (test &rest body)
   (if (and (consp body) (null (cdr body)))
      `(if ,test ,(car body))
      `(if ,test (progn ,@body))))

(defmacro unless (test &rest body)
   (if (and (consp body) (null (cdr body)))
      `(if ,test () ,(car body))
      `(if ,test () (progn ,@body))))

(defmacro dotimes (%loop &rest %body)
   ;; (DOTIMES (var count [res]) exp1 .. expN)
   (if (constantp (cadr %loop))
       `(FOR ((,(car %loop) 0 (1+ ,(car %loop))))
             ((>= ,(car %loop) ,(cadr %loop))
              ,(if (consp (cddr %loop)) (caddr %loop)))
             ,@%body)
       (let ((g (gensym)))
            `(FOR ((,(car %loop) 0 (1+ ,(car %loop)))
                   (,g ,(cadr %loop)))
                  ((>= ,(car %loop) ,g)
                   ,(if (consp (cddr %loop)) (caddr %loop)))
                  ,@%body))))

(defmacro dolist (%loop &rest %body)
   ;; (DOLIST (var listform [res]) exp1 .. expN)
   (let ((%dolist-var (gensym)))
        `(FOR ((,%dolist-var ,(cadr %loop) (cdr ,%dolist-var))
               (,(car %loop) ()))
              ((NULL ,%dolist-var) ,(if (consp (cddr %loop)) (caddr %loop)))
              (SETQ ,(car %loop) (car ,%dolist-var))
              ,@%body)))

(defmacro until (test &rest body)
  `(progn (while (null ,test) ,@body) t))

(defmacro repeat (n &rest body)
   (let ((repeat (gensym)))
       `(for ((,repeat (1- ,n) (1- ,repeat)))
             ((minusp ,repeat) t)
             ,@body)))

;;;
;;; Environment
;;;

(synonym 'defvar 'defdynamic)

(defmacro defkeyword (l)
   `(define-keyword ',l))

(defun get-function-name (fn)
   ;; returns the function name (if any) associated to a function object.
   (unless (functionp fn)
           (error "get-function-name: ~a is not a function.~%"))
   (or
       (some (lambda (x)
                (and (fboundp x)
                     (eq (symbol-function x) fn)
                     x))
             (oblist))
       (and (eq (function-type fn) 'expr)
            'lambda)))

;;;
;;; Packages
;;;

(defmacro defpackage (name &rest options)
   ;; The defpackage macro.
   (unless (or (symbolp name) (stringp name))
           (error "DEFPACKAGE: invalid package name ~a~%" name))
   (let ((exp       nil)
         (p         nil)
         (nicknames nil)
         (l         nil))
        (setq name (string name))
        (dolist (x options)
                (if (not (consp x))
                    (error "DEFPACKAGE: invalid list option ~a~%" x))
                (case (car x)
                      ((:SIZE)
                       ;; Ignore size.
                       (assure <integer> (cadr x)))
                      ((:SHADOW)
                       ;; Symbols   with   the  specified  names  are
                       ;; created  as  shadows  in  the package being
                       ;; defined.
                       (if (atom (cdr x))
                           (setq l (list (cdr x)))
                           (setq l (cdr x)))
                       (push `(SHADOW ',(mapcar #'string l) ,name) exp))
                      ((:USE)
                       ;; The  package being defined is made to "use"
                       ;; the packages specified.
                       (push 
                             `(USE-PACKAGE ',(mapcar #'string (cdr x)) ,name)
                              exp))
                      ((:IMPORT-FROM)
                       ;; Symbols   with   the  specified  names  are
                       ;; located  in  the  specified package.  These
                       ;; symbols   are  imported  into  the  package
                       ;; being defined.
                       (setq p (string (cadr x)))
                       (if (atom (cddr x))
                           (setq l (list (cddr x)))
                           (setq l (cddr x)))
                       (push
                             `(IMPORT (LIST ,@(mapcar (lambda (s)
                                                        `(INTERN ,s ,p))
                                                      (mapcar #'string l)))
                                      ,name)
                              exp))
                      ((:NICKNAMES)
                       ;; The  specified  names  become  nicknames of
                       ;; the package being defined.
                       (setf nicknames (cdr x)))
                      ((:INTERN)
                       ;; Symbols   with   the  specified  names  are
                       ;; located  or  created  in  the package being
                       ;; defined, just as with the function intern.
                       (dolist (s (cdr x))
                               (push `(INTERN ,(string s) ,name) exp)))
                      ((:EXPORT)
                       ;; Symbols   with   the  specified  names  are
                       ;; located  or  created  in  the package being
                       ;; defined  and  then  exported,  just as with
                       ;; the function export.
;                      (push
;                            `(EXPORT (LIST ,@(mapcar (lambda (s)
;                                                       `(INTERN ,s ,p))
;                                                     (mapcar #'string l)))
;                                     ,name)
;                             exp))
                       (push `(EXPORT ',(cdr x) ,name) exp))
                      (t
                       (error "DEFPACKAGE: invalid option ~a~%" x))))
        `(PROGN (CREATE-PACKAGE ,name ',nicknames)
                ,@(nreverse exp))))

(defmacro do-symbols (loop &rest body)
   ;; (DO-SYMBOLS (var [init [res]]) exp1 .. expN)
   ;; do-symbols  provides straightforward iteration over the symbols
   ;; of  a  package.  The  body  is  performed  once for each symbol
   ;; accessible  in  the package,  in no particular order,  with the
   ;; variable  var  bound to the symbol.  Then result-form (a single
   ;; form,  not  an implicit progn) is evaluated,  and the result is
   ;; the value of the do-symbols form.
   (let ((var (nth 0 loop))
         (pkg (nth 1 loop))
         (res (nth 2 loop))
         (it  (gensym)))
       `(for ((,it ,(if pkg `(oblist ',pkg) '(oblist)) (cdr ,it))
              (,var () ()))
             ((null ,it) ,res)
             (setq ,var (car ,it))
             ,@body)))

(defun libload (file &rest verbose)
   ;; load a library file (from *system-path*)
   (setq file (convert file <string>))
   (let ((f (search-in-path file)))
        (if f
            (load f verbose)
            (error "LIBLOAD: invalid filename '~a'~%" file))))

(set-macro-character #\control-l
     ;; to load file
     (lambda (stream char) `(libload ',(read-line stream))))

#-(eq (system-name) 'mvs)
(set-macro-character #\control-t
     ;; to launch a test file.
     (lambda (stream char)
        `(run-test ,(string-append (getenv "OPENLISP")
                                   "/tst/" (read-line stream) ".lsp"))))

;;;
;;; Missing sequence functions from standard lib
;;;

(defun remove-duplicates (seq)
   ;; This is a dispatcher for the correct algorithm.
   ;; Sequence is not physically modified.
   (cond
         ((consp seq)
          ;; Return a copy of sequence with redundant elements removed.
          (delete-duplicates-list (copy-seq seq)))
         ((or (vectorp seq) (stringp seq))
          (delete-duplicates-array seq))
         (t seq)))

(defun delete-duplicates (seq)
   ;; This is a dispatcher for the correct algorithm.
   (cond
         ((consp seq)
          (delete-duplicates-list seq))
         ((or (vectorp seq) (stringp seq))
          (delete-duplicates-array seq))
         (t seq)))

(defun delete-duplicates-list (list)
   ;; This is a specialized algorithm for lists.
   ;; Original list may be modified.
   (cond
         ((null list)
          nil)
         ((position (car list) (cdr list))
          (let ((res (delete-duplicates-list (cdr list))))
               (setf (car list) (car res))
               (setf (cdr list) (cdr res))))
         (t
          (setf (cdr list) (delete-duplicates-list (cdr list)))))
   list)

(defun delete-duplicates-array (seq)
   ;; This is a specialized algorithm for arrays.
   (cond
         ((> (length seq) 1)
          (let ((rest (subseq seq 1)))
               (cond
                     ((position (elt seq 0) rest)
                      (delete-duplicates-array rest))
                     (t
                      (append (subseq seq 0 1)
                              (delete-duplicates-array rest))))))
         (t seq)))

;;;
;;; Feature functions
;;;

(defun list-features ()
   *modules*)

(defun add-feature (package)
   (unless (featurep package)
           (unless (constantp package)
                   (set-symbol-global package package))
           (setq *modules* (cons (string package) *modules*))))

(defun rem-feature (package)
   (setq *modules* (remove (string package) *modules*)))

(defun featurep (package)
   (member (string package) *modules* :test #'equal))

;;;
;;; Standard functions (in ../lib/*)
;;;

(defglobal *autoload-required* ())

(defmacro autoload (file &rest fun)
   `(progn
           ,@(mapcar (lambda (f)
                       (setq *autoload-required*
                             (cons (cons f (cadr file)) *autoload-required*))
                       `(defmacro ,f (&rest body)
                           (require ,file)
                           (cons ',f body)))
                     fun)
           ',file))

(autoload 'defclass defclass)
(autoload 'flet     flet labels)
(autoload 'setf     setf defsetf incf decf push pop pushnew)

;;; Generic function for error reporting

(defmacro defgeneric (name arg &rest opt)
   ;; redefine  defgeneric  to  split  function  creation  and method
   ;; definition (one at a time).
   (let ((options ())
         (methods ()))
        (dolist (o opt)
                (case (car o)
                      ((:method)
                       (setq methods
                             (cons `(defmethod ,name ,@(cdr o)) methods)))
                      ((:method-combination)
                       (setq options (cons o options)))
                      ((:generic-function-class)
                       (setq options (cons o options)))
                      (t
                       (error "defmethod: unsupported option ~s." (car o)))))
        `(progn
                (make-generic-function ',name ',arg ',options)
                ,@methods)))

(defgeneric report-condition (condition stream)
   (:method ((condition <error>) (stream <stream>))
            (if (eq (elt condition 2) '<unbound-variable>)
                (format stream "** ~a : ~a~%"
                        (elt condition 1)
                        (elt condition 3))
                (format stream "** ~a : ~a : ~a~%"
                        (elt condition 1)
                        (elt condition 3)
                        (elt condition 2))))
   (:method ((condition <simple-error>) (stream <stream>))
            (apply #'format stream (elt condition 4) (elt condition 5))))

;;;
;;; Loader for standard environment.
;;;

(defun load-stdlib (compiled &rest verbose)
   ;; Load standard environment. Order is important if compiled = t.
   (let ((quiet *quiet-mode*))
        (setq *quiet-mode* t)
        (mapc (lambda (file)
                 (load (string-append (getenv "OPENLISP")
                                      "/lib/"
                                      (string file)
                                      (if compiled ".lap" ".lsp"))
                       verbose))
              '(startup
                opcode
                loader
                compiler
                setf
                fasl
                defstruc
                defclass
                datetime
                error))
        (setq *quiet-mode* quiet)))

;;;
;;; Shell argument $0, $1, .. $n. $# = arg. count.
;;;

(defglobal *command-shift* 0)

(defun argument-dispatcher (stream char)
   (let ((c (preview-char stream)))
        (identity char)
        (case c
              ((#\0)
               ;; program name.
               (read-byte stream)
              '(line-argument 0))
              ((#\*)
               ;; compute as a vector
               (read-byte stream)
              '(subseq (line-argument) (1+ *command-shift*)))
              ((#\@)
               ;; compute as a list
               (read-byte stream)
              '(convert (subseq (line-argument) (1+ *command-shift*))
               <list>))
              ((#\#)
               ;; compute the number of arguments
               (read-byte stream)
              `(- ,(1- (length (line-argument))) *command-shift*))
              ((#\$)
               ;; compute process ID may be 1000 if getpid is not available.
               (read-byte stream)
               (getpid))
              (t
               ;; returns nth argument
              `(line-argument (+ ,(read stream) *command-shift*))))))

(set-macro-character #\$ #'argument-dispatcher)

(defun shift ()
   ;; shift command line argument.
   (if (< *command-shift* (1- (length (line-argument))))
       (setq *command-shift* (1+ *command-shift*))
       (error "shift: no more argument.~%")))

;;;
;;; Extended format
;;;

(defglobal *format-table* (make-hash-table))

(defun add-format (type fn)
   (puthash type *format-table* fn))

(defun rem-format (type)
   (remhash type *format-table*))

(defun get-format (type)
   (gethash type *format-table*))

(defun format-user-type (obj st v)
   ;; when v (verbatim flag) is set, the function should print the object in
   ;; in a manner suitable for read operation.
   (let ((fn (gethash (type-of obj) *format-table* #'default-format-object)))
        (if (symbolp fn)
            (setq fn (symbol-function fn)))
        (if (not (functionp fn))
            (error "FORMAT-USER-TYPE: undefined function ~a" fn))
        (funcall fn obj st v))
   obj)

;;;
;;; Extented functions (in ./lib/)
;;;

(autoload 'apropos  apropos lhoblist)
(autoload 'calltree tc)
(autoload 'codegen  ccompile)
(autoload 'compress compress)
(autoload 'compiler compile compile-file cp)
(autoload 'datetime current-date make-date date= date/= date> date>= date< date<=)
(autoload 'defstruc defstruct make %build-class)
(autoload 'describe describe)
(autoload 'disasm   disassemble dump-lap)
(autoload 'extern   external-module)
(autoload 'fasl     fasl-dump)
(autoload 'gcstats  room)
(autoload 'loader   lap-loader lap-require)
(autoload 'logfile  log-format)
(autoload 'mine     mine mine-toplevel)
(autoload 'mvalues  values values-list multiple-values-setq)
(autoload 'pretty   pretty pprint)
(autoload 'profile  profile profile-all profile-log)
(autoload 'runtest  run-test)
(autoload 'sort     sort sortl sortn make-specialized-sort-function)
(autoload 'sysinfo  system-info)
(autoload 'trace    trace untrace)

;;;
;;; Try to load external modules
;;;

(unless *compiled-env*
        (let ((p (in-package "openlisp")))
             (external-modules t)
             (in-package p)))

;;;
;;; Tools and demos (in ./contrib/)
;;;

(unless (fboundp 'fib)
        (autoload 'fib fib))

(autoload 'hanoi    hanoi)

(set-macro-character #\control-v
     (lambda (stream char) (pretty (read stream)) t))

(unless (fboundp 'initty)
        (autoload 'virtty initty typrologue))

(when (fboundp 'edit)
      (defglobal *edit-name* "scratch")
      (add-feature 'edit))

(when (member (system-name) '(windows pocket-pc))
      (defun edit (file)
         (mine-toplevel (list file))))

#+(fboundp 'comline)
(progn

       (unless (fboundp 'edit)
               (defun edit (&rest file)
                  (when file
                        (setq *edit-name* (search-in-path (car file))))
                  (comline (string-append (or (getenv "EDITOR") "emacs")
                                          " "
                                          *edit-name*))
                  *edit-name*))

       (set-macro-character #\!
          (lambda (stream char)
             (let ((cmd (string-trim #(#\Space #\Tab) (read-line stream))))
                  (eq (comline cmd) 0))))

       (defmacro spawn (cmd &rest args)
          `(comline (format () ,cmd ,@args)))
)

#+(or (fboundp 'comline) (fboundp 'edit))
(progn
       (defglobal *edit-name* "scratch.lsp")

#-(eq (system-name) 'mvs)
       (set-macro-character #\control-e
            (lambda (stream char)
               (if (member (preview-char stream) '(#\newline #\carriage-return))
                   (progn
                          (read-char stream)
                          (edit *edit-name*))
                   (edit (setq *edit-name* (string (read stream)))))
               (format (standard-output) ";; Edito cum Emacis, ergo sum.~%")
               *edit-name*))

       (defun ed (&rest name)
          (when name (setq name (car name)))
          (cond
                ((null name)
                 (edit *edit-name*))
                ((stringp name)
                 (edit (setq *edit-name* name)))
                ((symbolp name)
                 (let ((file "scratch.lsp"))
                      (with-open-output-file (st file)
                            (with-standard-output st
                                  (format st ";; ~a definition :~2%" name)
                                  (pretty name)))
                      (edit (setq *edit-name* file))
                      (load file)
                      (delete-file file)))))
)

;;;
;;; Debugger
;;;

(defun debug (flag)
   ;; [de]activate the internal debugger on errors.
   (require 'pretty)
   (when (featurep 'lap)
         (require 'disasm))
   (internal-debug flag))

(defun break (msg &rest args)
   ;; display a message and directly goes to the debugger.
   (debug t)
   (apply #'cerror msg args))

;;;
;;; Socket
;;;

(defmacro with-client-socket (socket &rest body)
   `(let ((,(car socket) (socket)))
         (when ,(car socket)
               (unwind-protect (when ,`(connect ,@socket)
                                     ,@body)
                               (close ,(car socket))))))

(defmacro with-server-socket (socket &rest body)
   `(let ((,(car socket) (socket)))
         (when ,(car socket)
               (unwind-protect (when ,`(listen ,@socket)
                                     ,@body)
                               (close ,(car socket))))))

(when (fboundp 'dbexecute)
      (provide 'sql))

;;; Customization (those are default values)

(set-dynamic "? " *prompt*)                     ;; default prompt ?
(set-dynamic t    *read-level*)                 ;; print read level 2>
(set-dynamic 2    *warning-level*)              ;; OpenLisp warning level [0, 2]
(set-dynamic nil  *print-nil-as-list*)          ;; nil or ()
(set-dynamic 10   *print-base*)                 ;; integer default print base
(set-dynamic 256  *print-length*)               ;; max. list to print
(set-dynamic 256  *print-level*)                ;; max. recursive print level
(set-dynamic t    *print-stat*)                 ;; print evaluation time
(set-dynamic nil  *print-packages*)             ;; always print packages.
(set-dynamic nil  *continuable-system-error*)   ;; continuable errors

(when (featurep "emacs-pipe")
      ;; redefine some values when inside emacs.
      (set-dynamic "? " *prompt*)
      (set-dynamic nil  *read-level*)
      t)

(if (and (not *quiet-mode*) (not *compiled-env*))
    (format (standard-output)
            (if (member (system-name) '(pocket-pc))
                ";; OpenLisp is back again!~%"
                ";; God thank you, OpenLisp is back again!~%")))

(when (and (eq (function-type 'backconstant) 'subrn)
           (probe-file "autorun.lsp"))
      (load "autorun.lsp"))

;; (format (standard-output) ";; ~A~%" (list-features)))

(when (member (system-name) '(pocket-pc))
      (change-directory (getenv "OPENLISP")))

(in-package "user")

;;;
;;; End of startup bootstrap
;;;
