;;;; -*-Mode:LISP; Package:LISP; Base:10; Syntax:ISLISP -*-
;;;; Date:      2006/04/11
;;;; Title:     stderror.lsp
;:;; Author:    C. Jullien
;;;; RCS:       "$Id: stderror.lsp,v 1.3 2006/09/02 07:17:41 jullien Exp $"

;;; Standard conditions for OpenLisp

(require 'defclass)
(provide 'stderror)

(in-package "openlisp")

(defclass <serious-condition> ()
   ;; This  is  the  base  class for Conditions.  The following
   ;; slots  are  extensions  to ISLISP <SERIOUS-CONDITION> and
   ;; are  defined  there  for  compatibility with the internal
   ;; error  mechanism.  PLEASE  NOTE  that  order is important
   ;; since  the internal error handler also use a vector of at
   ;; least  4  arguments  which  can  be  passed  everywhere a
   ;; condition   object   is   expected.  A  compliant  ISLISP
   ;; application should never use those slots.
   ((serious-condition-continuable
       :initform nil
       :initarg  continuable
       :accessor serious-condition-continuable)
    (serious-condition-function
       :initform nil
       :initarg  function
       :accessor serious-condition-function)
    (serious-condition-arguments
       :initform nil
       :initarg  arguments
       :accessor serious-condition-arguments)
    (serious-condition-message
       :initform nil
       :initarg  message
       :accessor serious-condition-message)))

(defclass <storage-exhausted> (<serious-condition>)
   ())

(defclass <error> (<serious-condition>)
   ())

(defclass <simple-error> (<error>)
   ((simple-error-format-string
       :reader  simple-error-format-string
       :initarg format-string)
    (simple-error-format-arguments
       :reader  simple-error-format-arguments
       :initarg format-arguments)))

(defclass <parse-error> (<error>)
   ((parse-error-string
       :reader  parse-error-string
       :initarg string)
    (parse-error-expected-class
       :reader  parse-error-expected-class
       :initarg class)))

(defclass <control-error> (<error>)
   ())

(defclass <program-error> (<error>)
   ())

(defclass <stream-error> (<error>)
   ((stream-error-stream
       :reader  stream-error-stream
       :initarg stream)))

(defclass <arithmetic-error> (<error>)
   ((arithmetic-error-operation
       :reader  arithmetic-error-operation
       :initarg operation)
    (arithmetic-error-operands
       :reader  arithmetic-error-operands
       :initarg operands)))

(defclass <domain-error> (<program-error>)
   ((domain-error-object
       :reader  domain-error-object
       :initarg object)
    (domain-error-expected-class
       :reader  domain-error-expected-class
       :initarg expected-class)))

(defclass <undefined-entity> (<program-error>)
   ((undefined-entity-name
       :reader  undefined-entity-name
       :initarg name)
    (undefined-entity-namespace
       :reader  undefined-entity-namespace
       :initarg namespace)))

(defclass <unbound-variable> (<undefined-entity>)
   ())

(defclass <undefined-function> (<undefined-entity>)
   ())

(defclass <divide-by-zero> (<arithmetic-error>)
   ())

(defclass <floating-point-underflow> (<arithmetic-error>)
   ())

(defclass <floating-point-overflow> (<arithmetic-error>)
   ())

;;; Redefine error and cerror functions to use condition objects.

(defun error (string &rest obj)
   (signal-condition
           (create (class <simple-error>)
                   'format-string    string
                   'format-arguments obj)
           nil))

(defun cerror (continue-string error-string &rest obj)
   (signal-condition
           (create (class <simple-error>)
                   'format-string    error-string
                   'format-arguments obj)
           (let ((str (create-string-output-stream)))
                (format str continue-string obj)
                (get-output-stream-string str)
                (close str))))

;;; Generic function for error reporting

(defgeneric report-condition (condition stream)
   (:method ((condition <serious-condition>) (stream <stream>))
            (if (eq (serious-condition-arguments condition) '<unbound-variable>)
                (format stream "** ~a : ~a~%"
                        (serious-condition-function  condition)
                        (serious-condition-message   condition))
                (format stream "** ~a : ~a : ~a~%"
                        (serious-condition-function  condition)
                        (serious-condition-message   condition)
                        (serious-condition-arguments condition))))
   (:method ((condition <simple-error>) (stream <stream>))
            (apply #'format
                   stream
                   (simple-error-format-string    condition)
                   (simple-error-format-arguments condition)))
   (:method ((condition <domain-error>) (stream <stream>))
            (format stream "** ~a : domain-error, ~a expected : ~a~%"
                    (serious-condition-function  condition)
                    (domain-error-expected-class condition)
                    (domain-error-object         condition)))
   (:method ((condition <arithmetic-error>) (stream <stream>))
            (format stream "** ~a : arithmetic error : ~a~%"
                    (arithmetic-error-operation condition)
                    (arithmetic-error-operands  condition)))
   (:method ((condition <undefined-entity>) (stream <stream>))
            (format stream "** ~a : undefined entity : ~a~%"
                    (undefined-entity-name      condition)
                    (undefined-entity-namespace condition))))
